

def read_swap_trajectory ( fname ):

  inds,nswap,temps,enes = [],[],[],[]
  with open(fname, 'r') as f:
    for l in f.readlines():
      i,n,t,e = l.split()
      inds.append(int(i))
      nswap.append(int(n))
      temps.append(float(t))
      enes.append(float(e))

  return inds, nswap, temps, enes


def write_swap_accept ( fname,
                        write_mode,
                        step,
                        prev_accept,
                        temperature,
                        energy,
                        occupation_factors=None ):

  with open(fname, write_mode) as f:
    if occupation_factors is None:
      f.write(' '.join(map(str,[step, prev_accept, temperature, energy]))+'\n')



import numpy as np


def sps_swap ( species:list, nfixed=0, nswaps=1 ) -> dict:
  '''
    Interchange pairs in a list, stochastically

    Arguments:
      species (list): The list of elements to interchange
      nfixed (int): The number of elements to remain fixed, which must be placed at the beginning of the list
      nswaps (int): The number of swaps to perform

    Returns:
      (list): The resulting list with elements exchanged
  '''

  nat = len(species)
  t_species = species.copy()

  for _ in range(nswaps):

    ri1 = np.random.randint(nfixed, nat)
    ri2 = np.random.randint(nfixed, nat)

    while ri1 == ri2 or species[ri1] == species[ri2]:
      ri2 = np.random.randint(nfixed, nat)

    t_species[ri1],t_species[ri2] = t_species[ri2],t_species[ri1]

  return t_species



def sps_fixed ( lattice,
                species,
                positions,
                temperatures,
                temp_swaps=[],
                nfixed = 0,
                nswap = 2,
                nswap_inc = 1000,
                swap_fname = 'swaps.out',
                emin_fname = 'structure.emin.xyz',
                calculator = None,
                stop=None ):
  '''
    Perform the SPS routine on a fixed atomic basis without vacant sites.

    Arguments:
      lattice (list or ndarray): 3x3 matrix representing the three lattice vectors [R1, R2, R3]
      species (list): List of atomic symbols for each constituent site
      positions (list or ndarray): Nx3 matrix, with the crystal 3-coordinate for each of N atomic sites
      temperatures (list or ndarray): Temperature trajectory for simulated annealing trials
      temp_swaps (list or ndarray): Number of swaps to perform at each temperature in the trajectory. (Must contain one number for each provided temperature.
      nfixed (int): Number of sites to neglect from the swapping routine. Fixed sites must come first in the species and positions lists.
      nswap (int): Number of swaps to be performed at each step. This value is increased after nswap_inc rejected iterations.
      nswap_inc (int): Number of rejected trial configurations performed before increasing nswap.
      swap_fname (str): File name for the trajectory output file.
      emin_filename (str): File name for the minimum energy structure, output in the xyz format.
      calculator (str): Calculator for evaluating the structure energy.
      stop (float): Terminate the trial if the predicted structure energy is at or below the provided stop value.
  '''
  from .file_io import write_swap_accept
  from os.path import isfile
  from ase.io import write
  from ase import Atoms

  # Boltzmann const in eV
  kB = 1.68e-23/1.602e-19

  # Trajectory variables and structure information
  itr = 0
  last_swap_i = 0
  sswap = nswap
  nat = len(species)
  lattice = np.array(lattice)
  positions = np.array(positions)
  write_atoms = Atoms(species, positions=positions@lattice, cell=lattice, pbc=[1,1,1])

  # Verify that the input arrays have appropriate dimensions
  if len(lattice.shape) != 2 or not (lattice.shape[0] == 3 and lattice.shape[1] == 3):
    raise ValueError('Lattice shape must be (3,3)')
  if len(positions.shape) != 2 or positions.shape[0] != len(species) or positions.shape[1] != 3:
    raise ValueError('Positions shape must be (N,3), where N is the number of provided species')
  if nfixed > positions.shape[0]-2:
    raise ValueError(f'Cannot fix {nfixed} of {nat} sites. Decrease {nfixed} or provide more sites.')
  if len(set(species)) <= 1:
    raise ValueError('Species list must contain more than one type of species')

  # If the number of swaps per temperature is undefined, assign each to 1
  if len(temp_swaps) == 0:
    temp_swaps = np.ones(len(temperatures), dtype=np.short)
  if len(temperatures) != len(temp_swaps):
    raise ValueError('temperatures and temp_swaps must contain the same number of elements.')

  # If there is no provided calculator, initialize the default M3GNet calculator
  if calculator is None:
    from .calculators import MEGNet_Calculator
    calculator = MEGNet_Calculator()

  ene = emin = calculator.predict_formation_energy(lattice, species, positions)

  # Initialize the swaps output file. Exit if the file exists already.
  if isfile(swap_fname):
    raise FileExistsError(f'File {swap_fname} already exists. Will not overwrite.')
  write_swap_accept(swap_fname, 'w', itr, itr-last_swap_i, temperatures[0], ene)

  # Trajectory iteration
  for i,temp in enumerate(temperatures):
    for _ in range(temp_swaps[i]):
      itr += 1

      rswap = 1 + np.random.randint(np.min((nat-nfixed,nswap)))
      t_species = sps_swap(species, nfixed=nfixed, nswaps=rswap)
      t_ene = calculator.predict_formation_energy(lattice, t_species, positions)

      dE = t_ene - ene
      boltz = False if temp==0 else np.exp(-dE/(kB*temp)) > np.random.rand()

      # Accept condition
      if dE < 0 or boltz:
        write_swap_accept(swap_fname, 'a', itr, itr-last_swap_i, temp, t_ene)
        ene = t_ene
        nswap = sswap
        last_swap_i = itr
        species = t_species
        if ene < emin:
          emin = ene
          if emin_fname is not None:
            write_atoms.set_chemical_symbols(species)
            write(emin_fname, write_atoms)

      else:
        if (itr-last_swap_i) % nswap_inc == 0:
          nswap += 1

      if stop is not None:
        if ene <= stop:
          return



def sps_vacancy ( lattice,
                  vacancy_species,
                  vacancy_positions,
                  temperatures,
                  temp_swaps=[],
                  fixed_species=None,
                  fixed_positions=None,
                  nswap = 2,
                  nswap_inc = 1000,
                  swap_fname = 'swaps.out',
                  emin_fname = 'structure.emin.xyz',
                  occupation_factors=None,
                  calculator = None,
                  stop=None ):
  '''
    Perform the SPS routine on a set of atomic sites that contain some vacant sites.

    Arguments:
      lattice (list or ndarray): 3x3 matrix representing the three lattice vectors [R1, R2, R3]
      fixed_species (list): List of atomic symbols for sites that cannot participate in swapping
      fixed_positions (list or ndarray): Nx3 matrix, with the crystal 3-coordinate for each of N fixed atomic sites
      vacancy_species (list): List of atomic symbols for sites participating in the swaping
      vacancy_sites (list or ndarray): Nx3 matrix, with crystal 3-coordinate of sites that can be occupied or unoccupied by vacancy_species
      temperatures (list or ndarray): Temperature trajectory for simulated annealing trials
      temp_swaps (list or ndarray): Number of swaps to perform at each temperature in the trajectory. (Must contain one number for each provided temperature.
      nfixed (int): Number of sites to neglect from the swapping routine. Fixed sites must come first in the species and positions lists.
      nswap (int): Number of swaps to be performed at each step. This value is increased after nswap_inc rejected iterations.
      nswap_inc (int): Number of rejected trial configurations performed before increasing nswap.
      swap_fname (str): File name for the trajectory output file.
      emin_filename (str): File name for the minimum energy structure, output in the xyz format.
      calculator (str): Calculator for evaluating the structure energy.
      stop (float): Terminate the trial if the predicted structure energy is at or below the provided stop value.
  '''
  from .file_io import write_swap_accept
  from os.path import isfile
  from ase.io import write
  from ase import Atoms

  # Boltzmann const in eV
  kB = 1.68e-23/1.602e-19

  # Trajectory variables
  itr = 0
  last_swap_i = 0
  sswap = nswap

  # Set up fixed positions
  if None in [fixed_species, fixed_positions]:
    if fixed_species is not None or fixed_positions is not None:
      print('User must provide 1:1::species:position to fix. No fixed species or positions will be included.')
    fixed_species = []
    fixed_positions = np.empty((0,3), dtype=float)
  else:
    fixed_positions = np.array(fixed_positions)

  # Set up lattice, species, vacancy positions
  species = fixed_species + vacancy_species
  lattice = np.array(lattice)
  vacancy_positions = np.array(vacancy_positions)

  # Verify that the input arrays have appropriate dimensions
  if len(lattice.shape) != 2 or not (lattice.shape[0] == 3 and lattice.shape[1] == 3):
    raise ValueError('Lattice shape must be (3,3)')
  if len(fixed_positions.shape) != 2 or fixed_positions.shape[0] != len(fixed_species) or fixed_positions.shape[1] != 3:
    raise ValueError('Fixed positions shape must be (N,3), where N is the number of fixed species')
  if len(vacancy_positions.shape) != 2 or vacancy_positions.shape[0] < len(vacancy_species):
    msg = f'Vacancy positions shape must be (N,3), where N is greater than or equal to the number of vacancy species.'
    raise ValueError(msg)

  # Static array lengths
  nvspecies = len(vacancy_species)
  nfixed = fixed_positions.shape[0]
  nvsites = vacancy_positions.shape[0]

  # Set the initial vacancy occupation
  vacancy_indices = [i for i in range(nvspecies)]
  vacancy_occupation = {i:i for i in vacancy_indices}
  v_positions = [vacancy_positions[i] for i in vacancy_occupation]

  # Create Atoms object for output
  positions = np.concatenate([fixed_positions, v_positions])
  write_atoms = Atoms(species, positions=positions@lattice, cell=lattice, pbc=[1,1,1])

  # If there is no provided calculator, initialize the default M3GNet calculator
  if calculator is None:
    from .calculators import MEGNet_Calculator
    calculator = MEGNet_Calculator()

  # If the number of swaps per temperature is undefined, assign each to 1
  if len(temp_swaps) == 0:
    temp_swaps = np.ones(len(temperatures), dtype=np.short)
  if len(temperatures) != len(temp_swaps):
    raise ValueError('temperatures and temp_swaps must contain the same number of elements.')

  ene = emin = calculator.predict_formation_energy(lattice, species, positions)

  # Initialize the swaps output file. Exit if the file exists already.
  if isfile(swap_fname):
    raise FileExistsError(f'File {swap_fname} already exists. Will not overwrite.')
  write_swap_accept(swap_fname, 'w', itr, itr-last_swap_i, temperatures[0], ene, occupation_factors=occupation_factors)

  # Trajectory iteration
  for i,temp in enumerate(temperatures):
    for _ in range(temp_swaps[i]):
      itr += 1

      tvinds = vacancy_indices.copy()
      tvocc = vacancy_occupation.copy()

      # Perform stochastic swaps
      ns, nc = 0, 1+np.random.randint(nswap)
      ri = np.random.randint(nvspecies)
      while ns < nc:
        nri = np.random.randint(nvsites)
        if nri in tvocc:
          if vacancy_species[tvocc[nri]] == vacancy_species[ri]:
            continue
          tri = tvocc[nri]
          tvocc[tvinds[ri]],tvocc[nri] = tvocc[nri],tvocc[tvinds[ri]]
          tvinds[ri],tvinds[tri] = nri,tvinds[ri]
        else:
          del tvocc[tvinds[ri]]
          tvinds[ri] = nri
          tvocc[nri] = ri
        ns += 1

      # Update positions arrays
      v_positions = np.array([vacancy_positions[i] for i in tvinds])
      positions[nfixed:,:] = v_positions

      # Calculate energy and evaluate Metropolis condition
      t_ene = calculator.predict_formation_energy(lattice, species, positions)
      dE = t_ene - ene
      boltz = False if temp==0 else np.exp(-dE/(kB*temp)) > np.random.rand()
      if dE < 0 or boltz:
        write_swap_accept(swap_fname, 'a', itr, itr-last_swap_i, temp, t_ene, occupation_factors=occupation_factors)
        ene = t_ene
        nswap = sswap
        last_swap_i = itr
        vacancy_indices = tvinds
        vacancy_occupation = tvocc
        if ene < emin:
          emin = ene
          if emin_fname is not None:
            write_atoms.set_scaled_positions(positions)
            write(emin_fname, write_atoms)

      else:
        # Increase the number of swaps after nswap_inc rejections
        if (itr-last_swap_i) % nswap_inc == 0:
          nswap += 1

      # Halt if stop condition is met
      if stop is not None:
        if ene <= stop:
          return



from pymatgen.core.structure import Structure

class Calculator:

  def __init__ ( self ):
    pass

  def predict_formation_energy ( self, lattice, species, positions ):
    pass



class MEGNet_Calculator (Calculator):

  model = None

  def __init__ ( self, model='Eform_MP_2019', model_fname=None ):

    if model_fname is None:
      from megnet.utils.models import load_model
      self.model = load_model(model)

    else:
      from megnet.models import MEGNetModel
      self.model = MEGNetModel.from_file(model_fname)


  def predict_formation_energy ( self, lattice, species, positions ):
    pymatgen_struct = Structure(lattice, species, positions)
    return self.model.predict_structure(pymatgen_struct).ravel()[0]






def create_supercell ( lattice, positions, species=None, nsc=(2,2,2) ):
  '''
    Create a supercell lattice and set of atomic positions from a unit arrangement.
    If species are provided, they will be duplicated accordingly.

    Arguments:
      lattice (ndarray): 3x3 matrix with the lattice vectors
      positions (ndarray): Nx3 matrix containing the unit positions as 3-vectors
      species (list): List of species to duplicate. If None, only lattice and positions are replicated
      nsc (tuple, list, or ndarray): Number of cells to replicate in each direction (nx,ny,nz)

    Returns:
      (ndarray,ndarray): supercell lattice and positions, if species is not specified
      (ndarray,ndarray,list): supercell lattice, positions, and replicated species, if species is specified
  '''
  import numpy as np

  supercell_positions = []

  for i in range(nsc[0]):
    for j in range(nsc[1]):
      for k in range(nsc[2]):
        origin = np.array([i,j,k])
        supercell_positions += [pos+origin for pos in positions]

  supercell_lattice = lattice.copy()
  supercell_positions = np.array(supercell_positions)
  for i in range(3):
    supercell_lattice[i] *= nsc[i]
    supercell_positions[:,i] /= nsc[i]

  if species is not None:
    supercell_species = np.prod(nsc) * species
    return supercell_lattice,supercell_positions,supercell_species

  else:
    return supercell_lattice,supercell_positions




from setuptools import setup
import os

defs = os.path.join('src')

long_description = open('./README.md', 'r').read()

setup(name='MCSPS',
      version='0.1',
      description='A Python tool for performing site permutation search on fixed atomic geometries.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Frank T. Cerasoli',
      author_email='ftcerasoli@ucdavis.edu',
      platforms='Unix',
      url='https://github.com/nanotheorygroup/MC-SPS',
      packages=['MCSPS'],
      package_dir={'MCSPS':'src'},
      install_requires=['ase', 'numpy'],
      extra_requires=['matplotlib', 'megnet'],
      python_requires='>=3.7'
)





def plot_swap_trajectory ( file_pattern, temperature=True, min_energy=None, show=True ):
  from matplotlib.collections import LineCollection
  from matplotlib import colors as mpcol
  from matplotlib import pyplot as plt
  from glob import glob
  import numpy as np

  from .file_io import read_swap_trajectory

  fig,ax = plt.subplots(figsize=(6,4))
  fig.suptitle('GaAs Swap Trajectory')
  ax.set_xlabel('Iteration')
  ax.set_ylabel('Energy (eV/atom)')

  colors = ['black', 'gray']
  col_line = None
  if temperature:
    colors = [['blue','red'], ['lightblue','pink']]

  max_ind = -np.inf
  max_temp = -np.inf
  min_ene,max_ene = np.inf,-np.inf

  ach_min = []
  all_inds = []
  all_enes = []
  all_temps = []

  for fn in glob(file_pattern):

    inds,pinds,temps,enes = read_swap_trajectory(fn)
    mine,maxe = np.min(enes),np.max(enes)
    maxt = np.max(temps)
    maxi = np.max(inds)

    if maxi > max_ind:
      max_ind = maxi
    if mine < min_ene:
      min_ene = mine
    if maxe > max_ene:
      max_ene = maxe
    if maxt > max_temp:
      max_temp = maxt

    all_inds.append(inds)
    all_enes.append(enes)

    if min_energy is not None:
      t_mine = np.min(enes)
      if np.isclose(min_energy,t_mine,atol=1e-4) or t_mine < min_energy:
        ach_min.append(True)
      else:
        ach_min.append(False)

    if temperature:
      all_temps.append(temps)

  color_norm = plt.Normalize(0, max_temp)

  if min_energy is not None:
    ax.hlines(min_energy, 0, max_ind, linestyle='--', color='black')

  ordered_inds = list(range(len(all_inds)))
  if min_energy is not None:
    ordered_inds = [i for i,a in enumerate(ach_min) if not a]
    ordered_inds += [i for i,a in enumerate(ach_min) if a]

  for i in ordered_inds:

    inds = all_inds[i] + [max_ind]
    enes = all_enes[i] + [all_enes[i][-1]] 

    gcolors = colors[0]
    if min_energy is not None:
      if not ach_min[i]:
        gcolors = colors[1]

    # Color the segments by temperature
    if temperature:

      # Create the line segments
      points = np.array([inds,enes]).T.reshape(-1, 1, 2)
      segments = np.concatenate([points[:-1], points[1:]], axis=1)

      # Create the color mapping
      cmap = mpcol.LinearSegmentedColormap.from_list('', gcolors)

      lc = LineCollection(segments, cmap=cmap, norm=color_norm)
      lc.set_array(all_temps[i])
      col_line = ax.add_collection(lc)

    # Dont color the segments
    else:
      ax.plot(inds, enes, color=gcolors)

  # Set axis boundaries
  yoffset = 0.05 * (max_ene - min_ene)
  ax.set_xlim(0, max_ind*1.01)
  ax.set_ylim(min_ene-yoffset, max_ene+yoffset)

  if temperature:
    cbar = fig.colorbar(col_line)
    cbar.set_label('Temperature (K)')

  if show:
    plt.show()
    return None,None

  else:
    return fig,ax

